#include "cctk.h"

!$Id: auxvars.f90,v 1.4 2007/08/08 20:46:59 carlos Exp $

module auxvars
use params

implicit none
CONTAINS

!--------------------------------------------------------------------
!
! Define all the quantities that need discrete derivatives but 
! are not the gridfunctions that are evolved (combinations of them with 
! background quantities). In the trivial case of the wave equation we will
! use this variables for the fluxes, which are the variables themselves
! with a opposite sign
!--------------------------------------------------------------------
   subroutine define_aux_vars(v_pt, w_pt, u_pt, par)
   use GF
   
   CCTK_REAL, dimension(:), intent(IN)      :: u_pt   
   CCTK_REAL, dimension(:), intent(INOUT)   :: v_pt
   CCTK_REAL, dimension(:)                  :: w_pt
   CCTK_REAL, dimension(:)                  :: par

   !! local vars
   CCTK_INT                     :: i, j, k, m, nx, ny, nz
   CCTK_REAL                    :: x, y, z   
   logical, parameter           :: ltrace2 = .false.
   
   x = w_pt(H_XPHYS)
   y = w_pt(H_YPHYS)
   z = w_pt(H_ZPHYS)

   if (ltrace2) write(0,*)'DEFINE_AUX_VARS: begin'

   nx = nint(par(P_NX))
   ny = nint(par(P_NY))
   nz = nint(par(P_NZ))

   if (ltrace2) write(0,*)'DEFINE_AUX_VARS: nx,ny,nz',nx,ny,nz

   v_pt(H_FDPHI)   = u_pt(H_D4PHI)
   v_pt(H_FxD4PHI) = u_pt(H_D1PHI)
   v_pt(H_FyD4PHI) = u_pt(H_D2PHI)
   v_pt(H_FzD4PHI) = u_pt(H_D3PHI)
  
   return
 end subroutine define_aux_vars

  !--------------------------------------------------------------------
  !
  !
  !
  !--------------------------------------------------------------------

  subroutine SolvePrimVars(v, w, u, par)
    use GF

    type(gridfunction), dimension(NV) :: v
    type(gridfunction), dimension(NU) :: u
    type(gridfunction), dimension(NW) :: w
    CCTK_REAL, dimension(:)           :: par

  end subroutine SolvePrimVars


!----------------------------------------------------------------------
!  This routine is empty
!----------------------------------------------------------------------
   !-----------------------------------------------------------------
   !
   !-----------------------------------------------------------------
    subroutine calcrhs2(u_dot, u, v, w, dxu, dyu, dzu, dxv, dyv, dzv, &
         imask, par)

      use GF
      implicit none

      type(gridfunction), dimension(NU)   :: dxu, dyu, dzu
      type(gridfunction), dimension(NV)   :: dxv, dyv, dzv
      type(gridfunction), dimension(NU)   :: u, u_dot
      type(gridfunction), dimension(NW)   :: w
      type(gridfunction), dimension(NV)   :: v
      CCTK_REAL, dimension(:)             :: par
      CCTK_INT, dimension(:,:,:)          :: imask

      return
    end subroutine calcrhs2


   
 end module auxvars
