#include "cctk.h"
!__________________________________________________________
!
!   prim2char --> computing the characteristic variables 
!                  (eigenvectors) from the primitive ones
!   char2prim --> computing the original variables from 
!                 the eigenvectos
!___________________________________________________________
module charvars
  use params
  implicit none
  private
  public prim2char, char2prim
  
contains
  
  !#####################################################################################!
  subroutine prim2char(par,v,w,direction,u,w_in,w_out,w_0,T,uevolved,uexact,dxu,dyu,dzu,sources) 
   
    implicit none
   
    CCTK_REAL, dimension(:), intent(in) :: par
    CCTK_REAL, dimension(:), intent(in) :: w,v
    character(len=*), intent(in):: direction   
    CCTK_REAL, dimension(:), intent(in)  :: u,uevolved,uexact,dxu,dyu,dzu
    CCTK_REAL, dimension(:), intent(out) :: w_in, w_out
    CCTK_REAL, dimension(:), intent(out) :: sources
    CCTK_REAL, dimension(:), intent(out) :: w_0
    CCTK_REAL, dimension(3):: n_down
    CCTK_REAL, dimension(3,3) :: T
 
    CCTK_REAL:: x,y,z,nu1,nu2,nu3
    CCTK_INT :: k, i, j
    CCTK_REAL :: phi,pi,d1phi,d2phi,d3phi,dnphi

    x = w(H_XPHYS)
    y = w(H_YPHYS)
    z = w(H_ZPHYS)

    phi   = u(H_PHI)
    d1phi = u(H_D1PHI)
    d2phi = u(H_D2PHI)
    d3phi = u(H_D3PHI)
    pi    = u(H_D4PHI)
            
    !---------------------------------------------------------!

    call what_boundary(x,y,z,direction,n_down)

    ! n, index up 
    nu1 = n_down(1)
    nu2 = n_down(2)
    nu3 = n_down(3)

    dnphi = nu1*d1phi + nu2*d2phi + nu3*d3phi        
    
    w_in(1)  = pi - dnphi
    w_out(1) = pi + dnphi
    
    w_0(1) = d1phi - dnphi*n_down(1)
    w_0(2) = d2phi - dnphi*n_down(2)
    w_0(3) = d3phi - dnphi*n_down(3)
    w_0(4) = phi

  end subroutine prim2char
  
  !#####################################################################################!
  subroutine char2prim(par,v,w,direction,u,w_in,w_out,w_0,T,uevolved,uexact) 
    implicit none
     
    CCTK_REAL, dimension(:), intent(in) :: par
    CCTK_REAL, dimension(:), intent(in):: w,uevolved,uexact,v
    character(len=*), intent(in):: direction
    CCTK_REAL, dimension(:), intent(out) :: u
    CCTK_REAL, dimension(:), intent(in) :: w_in, w_out
    CCTK_REAL, dimension(:), intent(in) :: w_0
    CCTK_REAL, dimension(3) :: n_down
    CCTK_REAL, dimension(3,3) :: T

    CCTK_REAL:: x,y,z,nu1,nu2,nu3
    CCTK_REAL :: phi,pi,d1phi,d2phi,d3phi,dnphi        
    CCTK_INT :: k, i, j
    
    x = w(H_XPHYS)
    y = w(H_YPHYS)
    z = w(H_ZPHYS)
    
    !---------------------------------------------------------!
    
    call what_boundary(x,y,z,direction,n_down)

    ! n, index up 
    nu1 = n_down(1)
    nu2 = n_down(2)
    nu3 = n_down(3)

    pi    = 0.5d0*(w_in(1) + w_out(1))
    dnphi = 0.5d0*(-w_in(1) + w_out(1))        

    d1phi = w_0(1) + dnphi*n_down(1)
    d2phi = w_0(2) + dnphi*n_down(2)
    d3phi = w_0(3) + dnphi*n_down(3)        
    phi   = w_0(4)
    
    u(H_PHI) = phi
    u(H_D1PHI) = d1phi
    u(H_D2PHI) = d2phi
    u(H_D3PHI) = d3phi
    u(H_D4PHI) = pi
    

  end subroutine char2prim

  !#####################################################################################!
  subroutine what_boundary(x,y,z,direction,n_down)

    implicit none
    CCTK_REAL, intent(in) :: x,y,z
    character(len=*), intent(in):: direction
    CCTK_REAL, dimension(3), intent(out) :: n_down    

    !^^^^^^^^^^^^^^^^^^^^^  FACES ^^^^^^^^^^^^^^^^^^^^^^^^^^^^!
    choose_boundary : if (direction=="xmax") then
       n_down=(/1.,0.,0./)
    else if (direction=="xmin") then
       n_down=(/-1.,0.,0./)
    else if (direction=="ymax") then
       n_down=(/0.,1.,0./)
    else if (direction=="ymin") then
       n_down=(/0.,-1.,0./)
    else if (direction=="zmax") then
       n_down=(/0.,0.,1./) 
    else if (direction=="zmin") then
       n_down=(/0.,0.,-1./) 

       
       !^^^^^^^^^^^^^^^^^^^^^  EDGES ^^^^^^^^^^^^^^^^^^^^^^^^^^^^!
    else if (direction=="xmin_ymin") then
       n_down=(/-1.,-1.,0./)
    else if (direction=="xmin_ymax") then
       n_down=(/-1.,1.,0./)
    else if (direction=="xmin_zmin") then
       n_down=(/-1.,0.,-1./)
    else if (direction=="xmin_zmax") then
       n_down=(/-1.,0.,1./)
    else if (direction=="xmax_ymin") then
       n_down=(/1.,-1.,0./)
    else if (direction=="xmax_ymax") then
       n_down=(/1.,1.,0./)
    else if (direction=="xmax_zmin") then
       n_down=(/1.,0.,-1./)
    else if (direction=="xmax_zmax") then
       n_down=(/1.,0.,1./)
    else if (direction=="ymin_zmin") then
       n_down=(/0.,-1.,-1./)
    else if (direction=="ymin_zmax") then
       n_down=(/0.,-1.,1./)
    else if (direction=="ymax_zmin") then
       n_down=(/0.,1.,-1./)
    else if (direction=="ymax_zmax") then
       n_down=(/0.,1.,1./)

       
       !^^^^^^^^^^^^^^^^^^^^^  CORNERS ^^^^^^^^^^^^^^^^^^^^^^^^^!
       
    else if (direction=="xmin_ymin_zmin") then
       n_down=(/-1.,-1.,-1./)
    else if (direction=="xmin_ymin_zmax") then
       n_down=(/-1.,-1.,1./)
    else if (direction=="xmin_ymax_zmin") then
       n_down=(/-1.,1.,-1./)
    else if (direction=="xmin_ymax_zmax") then
       n_down=(/-1.,1.,1./)
    else if (direction=="xmax_ymin_zmin") then
       n_down=(/1.,-1.,-1./)
    else if (direction=="xmax_ymin_zmax") then
       n_down=(/1.,-1.,1./)
    else if (direction=="xmax_ymax_zmin") then
       n_down=(/1.,1.,-1./)
    else if (direction=="xmax_ymax_zmax") then
       n_down=(/1.,1.,1./)

    else 
       write(*,*) "******** Wrong direction passed to what_boundary subroutine *************"
       STOP
    end if choose_boundary

  end subroutine what_boundary

  !#####################################################################################!

end module charvars
