#include "cctk.h"

!----------------------------------------------------------------
!
! 5 types of initial data
! 
!_________________________________________________________________

module initial
use params
implicit none
CONTAINS
  subroutine initialdata(u0, u2, v, w,imask,par)
    use GF

    type(gridfunction), dimension(NU)    :: u0,u2
    type(gridfunction), dimension(NV)    :: v
    type(gridfunction), dimension(NW)    :: w
    CCTK_INT, dimension(:,:,:)           :: imask
    CCTK_REAL, dimension(:)              :: par

    !! local vars
    CCTK_REAL, dimension(:,:,:),pointer  :: xx, yy, zz
    CCTK_REAL               :: r,r0,delta_r,amp_phi,amp_phi_dot,costheta,dx
    CCTK_REAL               :: xpt(3), upt(NU)
    integer                 :: i,j,k,m, idtype, nx, ny, nz
    logical, parameter      :: ltrace  = .false.
    logical, parameter      :: ltrace2 = .false.


    xx => w(H_XPHYS)%d
    yy => w(H_YPHYS)%d
    zz => w(H_ZPHYS)%d

    if (ltrace) then
      write(0,*)'>>> INITIALDATA : x ',xx(1,1,1),xx(2,1,1),xx(3,1,1)
      write(0,*)'>>> INITIALDATA : y ',yy(1,1,1),yy(1,2,1),yy(1,3,1)
      write(0,*)'>>> INITIALDATA : z ',zz(1,1,1),zz(1,1,2),zz(1,1,3)
    end if

    do k=1,NU
       u0(k)%d = 0.0
    end do

    idtype = nint(par(P_IDTYPE))

    R0 = par(P_R0)
    delta_r = par(P_DELTA_R)
    amp_phi = par(P_AMP_PHI)
    amp_phi_dot = par(P_AMP_PHI_DOT)

    dx = par(P_DX)

    nx = nint(par(P_NX))
    ny = nint(par(P_NY))
    nz = nint(par(P_NZ))

    
    !>>>>>>>>>>>>>>>>>>>> TYPE 2 DATA: Compact support spherical shell <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    if (idtype .eq. 2) then
       do k=1,nz
          do j=1,ny
             do i=1,nx

                r = sqrt(xx(i,j,k)**2+yy(i,j,k)**2+zz(i,j,k)**2)
                if((r-r0)**2.le.delta_r**2.and.r.gt.0) then

                   u0(H_PHI)%d(i,j,k) = amp_phi &
                        *((r-r0)**2-delta_r**2)**4/delta_r**8/r
                   u0(H_D1PHI)%d(i,j,k) = amp_phi &
                        *((r-r0)**2-delta_r**2)**3*(7.*r**2-6.*r*r0+r0**2 &
                        +delta_r**2)*xx(i,j,k)/r**2/delta_r**8
                   u0(H_D2PHI)%d(i,j,k) = amp_phi &
                        *((r-r0)**2-delta_r**2)**3*(7.*r**2-6.*r*r0+r0**2 &
                        +delta_r**2)*yy(i,j,k)/r**2/delta_r**8
                   u0(H_D3PHI)%d(i,j,k) = amp_phi &
                        *((r-r0)**2-delta_r**2)**3*(7.*r**2-6.*r*r0+r0**2 &
                        +delta_r**2)*zz(i,j,k)/r**2/delta_r**8

                   u0(H_D4PHI)%d(i,j,k) = amp_phi_dot &
                        *((r-r0)**2-delta_r**2)**3*8.*(r0-r)/delta_r**8/r
                end if
             end do
          end do
       end do
    else
       write(0,*)'Unknown idtype= ',idtype
    end if

  end subroutine initialdata
    
 end module initial   
