#include "cctk.h"


module boundary_mine
   use params

   use GF

   contains 

  !--------------------------------------------------------------------
  !
  !
  !
  !--------------------------------------------------------------------
  subroutine outflow_boundary(u, w, imask, par)
    implicit   none
    type(gridfunction), dimension(NU), intent(inout) :: u
    type(gridfunction), dimension(NW) :: w
    CCTK_REAL, dimension(:)           :: par
    CCTK_INT, dimension(:,:,:)        :: imask

    ! local vars
    CCTK_INT                          :: nx, ny, nz, ng
    
    nx = nint(par(P_NX))  ! local size of arrays
    ny = nint(par(P_NY))
    nz = nint(par(P_NZ))
    ng = nint(par(P_GHOSTWIDTH))

    !call apply_outflow(u(H_D)%d,   nx, ny, nz, ng)
    !call apply_outflow(u(H_SX)%d,  nx, ny, nz, ng)
    !call apply_outflow(u(H_SY)%d,  nx, ny, nz, ng)
    !call apply_outflow(u(H_SZ)%d,  nx, ny, nz, ng)
    !call apply_outflow(u(H_TAU)%d, nx, ny, nz, ng)
    !call apply_outflow(u(H_BX)%d,  nx, ny, nz, ng)
    !call apply_outflow(u(H_BY)%d,  nx, ny, nz, ng)
    !call apply_outflow(u(H_BZ)%d,  nx, ny, nz, ng)
    !call apply_outflow(u(H_PSI)%d, nx, ny, nz, ng)

    return
  end subroutine outflow_boundary

  !--------------------------------------------------------------------
  !
  !
  !
  !--------------------------------------------------------------------
  subroutine apply_outflow(f, nx, ny, nz, ng)
    implicit   none
    CCTK_INT                       :: nx, ny, nz, ng
    CCTK_REAL, dimension(nx,ny,nz) :: f

    ! local vars
    CCTK_INT                       :: i, j, k

    ! xmin
    do k = 1, nz
      do j = 1, ny
        do i = 1, ng
          f(i,j,k) = f(ng+1,j,k)
        end do
      end do
    end do

    ! xmax
    do k = 1, nz
      do j = 1, ny
        do i = nx-ng+1, nx
          f(i,j,k) = f(nx-ng,j,k)
        end do
      end do
    end do

    ! ymin
    do k = 1, nz
      do i = 1, nx
        do j = 1, ng
          f(i,j,k) = f(i,ng+1,k)
        end do
      end do
    end do

    ! ymax
    do k = 1, nz
      do i = 1, nx
        do j = ny-ng+1, ny
          f(i,j,k) = f(i,ny-ng,k)
        end do
      end do
    end do

    ! zmin
    do j = 1, ny
      do i = 1, nx
        do k = 1, ng
          f(i,j,k) = f(i,j,ng+1)
        end do
      end do
    end do

    ! zmax
    do j = 1, ny
      do i = 1, nx
        do k = nz-ng+1, nz
          f(i,j,k) = f(i,j,nz-ng)
        end do
      end do
    end do

    return
  end subroutine apply_outflow
  !--------------------------------------------------------------------
  !
  !
  !
  !--------------------------------------------------------------------

    subroutine my_boundary(u,udot,w, imask, par)
    type(gridfunction), dimension(NU), intent(inout) :: u, udot
    type(gridfunction), dimension(NW) :: w
    CCTK_REAL, dimension(:)           :: par
    CCTK_INT, dimension(:,:,:)        :: imask

    ! local vars
    CCTK_REAL, dimension(:,:,:), pointer :: xx, yy, zz
    CCTK_REAL, DIMENSION(3) :: n  ! outward unit normal
    CCTK_INT :: imin,imax,jmin,jmax,kmin,kmax, IBBOX1, IBBOX2, IBBOX3,&
         IBBOX4, IBBOX5, IBBOX6, nx, ny, nz, k, i, j, l, NB    
    CCTK_INT :: idx, idy, idz, mask_stencil
    
    IBBOX1 = nint(par(P_BBOX1))
    IBBOX2 = nint(par(P_BBOX2))
    IBBOX3 = nint(par(P_BBOX3))
    IBBOX4 = nint(par(P_BBOX4))
    IBBOX5 = nint(par(P_BBOX5))
    IBBOX6 = nint(par(P_BBOX6)) 
   
    nx = int(par(P_NX))  ! local size of arrays
    ny = int(par(P_NY))
    nz = int(par(P_NZ))

    xx => w(H_XPHYS)%d
    yy => w(H_YPHYS)%d
    zz => w(H_ZPHYS)%d

    write(*,*) 'inside my_boundary'

  end subroutine my_boundary





end module boundary_mine
