!----------------------------------------------------------------------------
!
!     $Id: rhs_noconserv.f90,v 1.1.1.1 2004/02/10 19:12:06 steve Exp $
!
!     This subroutine discretizes the RHS of the scalar wave equation on a
!     Schwarzschild background.   The equations use Olivier/s interpolating
!     formulation, but the discretization does NOT conserve a discrete
!     energy.  This RHS is to be used in comparison with the standard
!     RHS, which does preserve a discrete energy, purposes only.
!
!
!----------------------------------------------------------------------------
#include "cctk.h"

module rhs_mine_noconserv
use params
use background_quant
implicit none

contains

      subroutine calcrhs_noconserv(RHS, u, v, dxu, dyu, dzu, &
     &                             dxv, dyv, dzv, w, par)
      implicit   none

      CCTK_REAL, dimension(NU), intent(out) :: RHS
      CCTK_REAL, dimension(NU), intent(in) :: u,dxu,dyu,dzu
      CCTK_REAL, dimension(NV), intent(in) :: dxv,dyv,dzv, v
      CCTK_REAL, dimension(NW), intent(in) :: w
      CCTK_REAL, dimension(NPAR),intent(in) :: par


!     Geometric variables
      CCTK_REAL x, y, z, xx(3)
      CCTK_REAL H, l(0:3), L0, L1, L2, L3, dH(0:3), dl(0:3,0:3)
      CCTK_REAL gd(3,3), gxx, gxy, gxz, gyy, gyz, gzz
      CCTK_REAL gu(3,3), guxx, guxy, guxz, guyy, guyz, guzz
      CCTK_REAL alpha, betadx, betady, betadz, dalpha(3)
      CCTK_REAL dxalpha, dyalpha, dzalpha
      CCTK_REAL dbetad(3,3), dgd(3,3,0:3), dgu(3,3,3), Bu(3), dBu(3,3)
      CCTK_REAL betau(3), dbetau(3,3)
      CCTK_REAL dxbetadx, dybetadx, dzbetadx
      CCTK_REAL dxbetady, dybetady, dzbetady
      CCTK_REAL dxbetadz, dybetadz, dzbetadz
      CCTK_REAL dtgxx, dxgxx, dygxx, dzgxx
      CCTK_REAL dtgxy, dxgxy, dygxy, dzgxy
      CCTK_REAL dtgxz, dxgxz, dygxz, dzgxz
      CCTK_REAL dtgyy, dxgyy, dygyy, dzgyy
      CCTK_REAL dtgyz, dxgyz, dygyz, dzgyz
      CCTK_REAL dtgzz, dxgzz, dygzz, dzgzz
      CCTK_REAL dxguxx, dyguxx, dzguxx, dxguxy, dyguxy, dzguxy
      CCTK_REAL dxguxz, dyguxz, dzguxz, dxguyy, dyguyy, dzguyy
      CCTK_REAL dxguyz, dyguyz, dzguyz, dxguzz, dyguzz, dzguzz
      CCTK_REAL betax, betay, betaz
      CCTK_REAL dxbetax, dybetax, dzbetax
      CCTK_REAL dxbetay, dybetay, dzbetay
      CCTK_REAL dxbetaz, dybetaz, dzbetaz
      CCTK_REAL c111, c112, c113, c122, c123, c133, c211, c212, c213
      CCTK_REAL c222, c223, c233, c311, c312, c313, c322, c323, c333
      CCTK_REAL trK, Bx, By, Bz
      CCTK_REAL dxBx, dyBx, dzBx
      CCTK_REAL dxBy, dyBy, dzBy
      CCTK_REAL dxBz, dyBz, dzBz
      CCTK_REAL M

!     Scalar field varialbes
      CCTK_REAL PHI, PI, P1, P2, P3, Vsf
      CCTK_REAL dxPHI, dyPHI, dzPHI
      CCTK_REAL dxPI, dyPI, dzPI, dxP1, dyP1, dzP1, dxP2, dyP2, dzP2
      CCTK_REAL dxP3, dyP3, dzP3

!     Maple vars
      CCTK_INT  i
      CCTK_REAL  t1, t2, t3, t5, t6
      CCTK_REAL  t7, t9, t8, t10
      CCTK_REAL  t4, t11, t13, t15, t19
      CCTK_REAL  t21, t27, t36, t16, t20
      CCTK_REAL  t23, t24, t29, t30, t33
      CCTK_REAL  t38, t39, t42, t46, t49
      CCTK_REAL  t63, t66, t17, t12, t14
      CCTK_REAL  t18, t25, t, t28, t31
      CCTK_REAL  t34, t35, t37, t40, t41
      CCTK_REAL  t43, t50, t54, t58, t62
      CCTK_REAL  t90, t96, t102, t108, t109
      CCTK_REAL  t112, t22, t26, t32, t47
      CCTK_REAL  t56, t60, t64, t67, t70
      CCTK_REAL  t71, t73, t77, t80, t81
      CCTK_REAL  t83, t89, t92, t98, t99
      CCTK_REAL  t104, t105, t107, t110, t114
      CCTK_REAL  t117, t119, t121, t125, t142
      CCTK_REAL  t149, t156, t160, s1, t87
      CCTK_REAL  t91, t93, t94, t100, t145
      CCTK_REAL  t146, t173, t220


      ! The potential for the Scalar field appears in the equations
      ! as d/dPHI V(PHI) =: Vsf
      Vsf = 0.0d0

      M   =  par(P_BH1_MASS)

      x = w(H_XPHYS)
      y = w(H_YPHYS)
      z = w(H_ZPHYS)

      xx(1) = x
      xx(2) = y
      xx(3) = z

      if (x .eq. 0.0d0 .and. y .eq. 0.0d0 .and. z .eq. 0.0d0) then
        !write(0,*)'calling RHS at origin '
        do i = 1, NU
          RHS(i) = 0.0d0
        end do
        return
      end if

      PI  = u(H_D4PHI)
      PHI = u(H_PHI)
      P1  = u(H_D1PHI)
      P2  = u(H_D2PHI)
      P3  = u(H_D3PHI)

      dxPHI = dxu(H_PHI)
      dxP1  = dxu(H_D1PHI)
      dxP2  = dxu(H_D2PHI)
      dxP3  = dxu(H_D3PHI)
      dxPI  = dxu(H_D4PHI)

      dyPHI = dyu(H_PHI)
      dyP1  = dyu(H_D1PHI)
      dyP2  = dyu(H_D2PHI)
      dyP3  = dyu(H_D3PHI)
      dyPI  = dyu(H_D4PHI)

      dzPHI = dzu(H_PHI)
      dzP1  = dzu(H_D1PHI)
      dzP2  = dzu(H_D2PHI)
      dzP3  = dzu(H_D3PHI)
      dzPI  = dzu(H_D4PHI)

       
      t1 = x**2
      t2 = y**2
      t3 = z**2
      t5 = (t1+t2+t3)**2
      t6 = sqrt(t5)
      t7 = t1+t2+t3+t6
      t9 = sqrt(2.D0)*sqrt(t7)
      H = M*t9/t7
      t1 = x**2
      t2 = y**2
      t3 = z**2
      t5 = (t1+t2+t3)**2
      t6 = sqrt(t5)
      t7 = t1+t2+t3+t6
      t8 = sqrt(2*t7)
      t10 = 2/t7
      l(0) = 1
      l(1) = t8*x*t10/2
      l(2) = t8*y*t10/2
      l(3) = 2*z/t8
      L0=l(0)
      L1=l(1)
      L2=l(2)
      L3=l(3)
      t1 = x**2
      t2 = y**2
      t3 = z**2
      t4 = t1+t2+t3
      t5 = t4**2
      t6 = sqrt(t5)
      t7 = t1+t2+t3+t6
      t8 = sqrt(2.0d0*t7)
      t10 = M/t8
      t11 = 2/t7
      t13 = 1/t6*t4
      t15 = x+t13*x
      t19 = M*t8
      t21 = 4/t7**2
      t27 = y+t13*y
      t36 = z+t13*z
      dH(0) = 0
      dH(1) = t10*t11*t15-t19*t21*t15/2.0d0
      dH(2) = t10*t11*t27-t19*t21*t27/2.0d0
      dH(3) = t10*t11*t36-t19*t21*t36/2.0d0
      t1 = x**2
      t2 = y**2
      t3 = z**2
      t4 = t1+t2+t3
      t5 = t4**2
      t6 = sqrt(t5)
      t7 = t1+t2+t3+t6
      t8 = sqrt(2*t7)
      t9 = 1/t8
      t10 = t9*x
      t11 = 2/t7
      t13 = 1/t6*t4
      t15 = x+t13*x
      t16 = 4.0d0*t11*t15
      t20 = t8*t11/2.0d0
      t21 = t8*x
      t23 = 4/t7**2
      t24 = t23*t15
      t29 = y+t13*y
      t30 = 4*t11*t29
      t33 = t23*t29
      t38 = z+t13*z
      t39 = 4*t11*t38
      t42 = t23*t38
      t46 = t9*y
      t49 = t8*y
      t63 = t8**2
      t66 = z/t63/t8
      dl(0,0) = 0.0d0
      dl(0,1) = 0.0d0
      dl(0,2) = 0.0d0
      dl(0,3) = 0.0d0
      dl(1,0) = 0.0d0
      dl(1,1) = t10*t16/4.0d0+t20-t21*t24/2.0d0
      dl(1,2) = t10*t30/4.0d0-t21*t33/2.0d0
      dl(1,3) = t10*t39/4.0d0-t21*t42/2.0d0
      dl(2,0) = 0.0d0
      dl(2,1) = t46*t16/4.0d0-t49*t24/2.0d0
      dl(2,2) = t46*t30/4.0d0+t20-t49*t33/2.0d0
      dl(2,3) = t46*t39/4.0d0-t49*t42/2.0d0
      dl(3,0) = 0.0d0
      dl(3,1) = -4*t66*t15
      dl(3,2) = -4*t66*t29
      dl(3,3) = 2*t9-4*t66*t38
      t1 = L1**2
      t5 = H*L1
      t7 = 2*t5*L2
      t9 = 2*t5*L3
      t10 = L2**2
      t16 = 2*H*L2*L3
      t17 = L3**2
      gd(1,1) = 1+2*H*t1
      gd(1,2) = t7
      gd(1,3) = t9
      gd(2,1) = t7
      gd(2,2) = 1+2*H*t10
      gd(2,3) = t16
      gd(3,1) = t9
      gd(3,2) = t16
      gd(3,3) = 1+2*H*t17
      gxx=gd(1,1)
      gxy=gd(1,2)
      gxz=gd(1,3)
      gyy=gd(2,2)
      gyz=gd(2,3)
      gzz=gd(3,3)
      t1 = L3**2
      t3 = 2*H*t1
      t4 = L2**2
      t6 = 2*H*t4
      t8 = L1**2
      t10 = 2*H*t8
      t12 = 1/(1+t10+t6+t3)
      t14 = H*L1
      t17 = 2*t14*L2*t12
      t18 = L3*t12
      t20 = 2*t14*t18
      t25 = 2*H*L2*t18
      gu(1,1) = (1+t3+t6)*t12
      gu(1,2) = -t17
      gu(1,3) = -t20
      gu(2,1) = -t17
      gu(2,2) = (1+t3+t10)*t12
      gu(2,3) = -t25
      gu(3,1) = -t20
      gu(3,2) = -t25
      gu(3,3) = (1+t6+t10)*t12
      guxx=gu(1,1)
      guxy=gu(1,2)
      guxz=gu(1,3)
      guyy=gu(2,2)
      guyz=gu(2,3)
      guzz=gu(3,3)
      t1 = L1**2
      t2 = L2**2
      t3 = L3**2
      t8 = sqrt(1+2*H*(t1+t2+t3))
      alpha = 1/t8
      betadx = 2*H*L0*L1
      betady = 2*H*L0*L2
      betadz = 2*H*L0*L3
      t1 = L1**2
      t2 = L2**2
      t3 = L3**2
      t4 = t1+t2+t3
      t8 = sqrt(1+2*H*t4)
      t9 = t8**2
      t11 = 1/t9/t8
      dalpha(1) = -t11*(dH(1)*t4+2*H*(L1*dl(1,1)+L2*dl(2,1)+L3*dl(3,1)))
      dalpha(2) = -t11*(dH(2)*t4+2*H*(L1*dl(1,2)+L2*dl(2,2)+L3*dl(3,2)))
      dalpha(3) = -t11*(dH(3)*t4+2*H*(L1*dl(1,3)+L2*dl(2,3)+L3*dl(3,3)))
      dxalpha=dalpha(1)
      dyalpha=dalpha(2)
      dzalpha=dalpha(3)
      t1 = dH(1)*L0
      t3 = H*dl(0,1)
      t5 = H*L0
      t8 = dH(2)*L0
      t10 = H*dl(0,2)
      t14 = dH(3)*L0
      t16 = H*dl(0,3)
      dbetad(1,1) = 2*t1*L1+2*t3*L1+2*t5*dl(1,1)
      dbetad(1,2) = 2*t8*L1+2*t10*L1+2*t5*dl(1,2)
      dbetad(1,3) = 2*t14*L1+2*t16*L1+2*t5*dl(1,3)
      dbetad(2,1) = 2*t1*L2+2*t3*L2+2*t5*dl(2,1)
      dbetad(2,2) = 2*t8*L2+2*t10*L2+2*t5*dl(2,2)
      dbetad(2,3) = 2*t14*L2+2*t16*L2+2*t5*dl(2,3)
      dbetad(3,1) = 2*t1*L3+2*t3*L3+2*t5*dl(3,1)
      dbetad(3,2) = 2*t8*L3+2*t10*L3+2*t5*dl(3,2)
      dbetad(3,3) = 2*t14*L3+2*t16*L3+2*t5*dl(3,3)
      dxbetadx=dbetad(1,1)
      dybetadx=dbetad(1,2)
      dzbetadx=dbetad(1,3)
      dxbetady=dbetad(2,1)
      dybetady=dbetad(2,2)
      dzbetady=dbetad(2,3)
      dxbetadz=dbetad(3,1)
      dybetadz=dbetad(3,2)
      dzbetadz=dbetad(3,3)
      t1 = L1**2
      t4 = H*L1
      t23 = dH(0)*L1
      t25 = H*dl(1,0)
      t28 = t23*L2+t25*L2+t4*dl(2,0)
      t29 = dH(1)*L1
      t31 = H*dl(1,1)
      t34 = t29*L2+t31*L2+t4*dl(2,1)
      t35 = dH(2)*L1
      t37 = H*dl(1,2)
      t40 = t35*L2+t37*L2+t4*dl(2,2)
      t41 = dH(3)*L1
      t43 = H*dl(1,3)
      t46 = t41*L2+t43*L2+t4*dl(2,3)
      t50 = t23*L3+t25*L3+t4*dl(3,0)
      t54 = t29*L3+t31*L3+t4*dl(3,1)
      t58 = t35*L3+t37*L3+t4*dl(3,2)
      t62 = t41*L3+t43*L3+t4*dl(3,3)
      t63 = L2**2
      t66 = H*L2
      t90 = dH(0)*L2*L3+H*dl(2,0)*L3+t66*dl(3,0)
      t96 = dH(1)*L2*L3+H*dl(2,1)*L3+t66*dl(3,1)
      t102 = dH(2)*L2*L3+H*dl(2,2)*L3+t66*dl(3,2)
      t108 = dH(3)*L2*L3+H*dl(2,3)*L3+t66*dl(3,3)
      t109 = L3**2
      t112 = H*L3
      dgd(1,1,0) = 2*dH(0)*t1+4*t4*dl(1,0)
      dgd(1,1,1) = 2*dH(1)*t1+4*t4*dl(1,1)
      dgd(1,1,2) = 2*dH(2)*t1+4*t4*dl(1,2)
      dgd(1,1,3) = 2*dH(3)*t1+4*t4*dl(1,3)
      dgd(1,2,0) = 2*t28
      dgd(1,2,1) = 2*t34
      dgd(1,2,2) = 2*t40
      dgd(1,2,3) = 2*t46
      dgd(1,3,0) = 2*t50
      dgd(1,3,1) = 2*t54
      dgd(1,3,2) = 2*t58
      dgd(1,3,3) = 2*t62
      dgd(2,1,0) = 2*t28
      dgd(2,1,1) = 2*t34
      dgd(2,1,2) = 2*t40
      dgd(2,1,3) = 2*t46
      dgd(2,2,0) = 2*dH(0)*t63+4*t66*dl(2,0)
      dgd(2,2,1) = 2*dH(1)*t63+4*t66*dl(2,1)
      dgd(2,2,2) = 2*dH(2)*t63+4*t66*dl(2,2)
      dgd(2,2,3) = 2*dH(3)*t63+4*t66*dl(2,3)
      dgd(2,3,0) = 2*t90
      dgd(2,3,1) = 2*t96
      dgd(2,3,2) = 2*t102
      dgd(2,3,3) = 2*t108
      dgd(3,1,0) = 2*t50
      dgd(3,1,1) = 2*t54
      dgd(3,1,2) = 2*t58
      dgd(3,1,3) = 2*t62
      dgd(3,2,0) = 2*t90
      dgd(3,2,1) = 2*t96
      dgd(3,2,2) = 2*t102
      dgd(3,2,3) = 2*t108
      dgd(3,3,0) = 2*dH(0)*t109+4*t112*dl(3,0)
      dgd(3,3,1) = 2*dH(1)*t109+4*t112*dl(3,1)
      dgd(3,3,2) = 2*dH(2)*t109+4*t112*dl(3,2)
      dgd(3,3,3) = 2*dH(3)*t109+4*t112*dl(3,3)
      dtgxx=dgd(1,1,0)
      dxgxx=dgd(1,1,1)
      dygxx=dgd(1,1,2)
      dzgxx=dgd(1,1,3)
      dtgxy=dgd(1,2,0)
      dxgxy=dgd(1,2,1)
      dygxy=dgd(1,2,2)
      dzgxy=dgd(1,2,3)
      dtgxz=dgd(1,3,0)
      dxgxz=dgd(1,3,1)
      dygxz=dgd(1,3,2)
      dzgxz=dgd(1,3,3)
      dtgyy=dgd(2,2,0)
      dxgyy=dgd(2,2,1)
      dygyy=dgd(2,2,2)
      dzgyy=dgd(2,2,3)
      dtgyz=dgd(2,3,0)
      dxgyz=dgd(2,3,1)
      dygyz=dgd(2,3,2)
      dzgyz=dgd(2,3,3)
      dtgzz=dgd(3,3,0)
      dxgzz=dgd(3,3,1)
      dygzz=dgd(3,3,2)
      dzgzz=dgd(3,3,3)
      t1 = L3**2
      t3 = 2*dH(1)*t1
      t4 = H*L3
      t6 = 4*t4*dl(3,1)
      t7 = L2**2
      t9 = 2*dH(1)*t7
      t10 = H*L2
      t12 = 4*t10*dl(2,1)
      t14 = L1**2
      t16 = 2*H*t14
      t18 = 2*H*t7
      t20 = 2*H*t1
      t21 = 1+t16+t18+t20
      t22 = 1/t21
      t25 = t21**2
      t26 = 1/t25
      t27 = (1+t20+t18)*t26
      t29 = 2*dH(1)*t14
      t30 = H*L1
      t32 = 4*t30*dl(1,1)
      t33 = t29+t32+t9+t12+t3+t6
      t37 = 2*dH(2)*t1
      t39 = 4*t4*dl(3,2)
      t41 = 2*dH(2)*t7
      t43 = 4*t10*dl(2,2)
      t47 = 2*dH(2)*t14
      t49 = 4*t30*dl(1,2)
      t50 = t47+t49+t41+t43+t37+t39
      t54 = 2*dH(3)*t1
      t56 = 4*t4*dl(3,3)
      t58 = 2*dH(3)*t7
      t60 = 4*t10*dl(2,3)
      t64 = 2*dH(3)*t14
      t66 = 4*t30*dl(1,3)
      t67 = t64+t66+t58+t60+t54+t56
      t70 = dH(1)*L1
      t71 = L2*t22
      t73 = H*dl(1,1)
      t77 = L2*t26
      t80 = -t70*t71-t73*t71-t30*dl(2,1)*t22+t30*t77*t33
      t81 = dH(2)*L1
      t83 = H*dl(1,2)
      t89 = -t81*t71-t83*t71-t30*dl(2,2)*t22+t30*t77*t50
      t90 = dH(3)*L1
      t92 = H*dl(1,3)
      t98 = -t90*t71-t92*t71-t30*dl(2,3)*t22+t30*t77*t67
      t99 = L3*t22
      t102 = dl(3,1)*t22
      t104 = L3*t26
      t105 = t104*t33
      t107 = -t70*t99-t73*t99-t30*t102+t30*t105
      t110 = dl(3,2)*t22
      t112 = t104*t50
      t114 = -t81*t99-t83*t99-t30*t110+t30*t112
      t117 = dl(3,3)*t22
      t119 = t104*t67
      t121 = -t90*t99-t92*t99-t30*t117+t30*t119
      t125 = (1+t20+t16)*t26
      t142 = -dH(1)*L2*t99-H*dl(2,1)*t99-t10*t102+t10*t105
      t149 = -dH(2)*L2*t99-H*dl(2,2)*t99-t10*t110+t10*t112
      t156 = -dH(3)*L2*t99-H*dl(2,3)*t99-t10*t117+t10*t119
      t160 = (1+t18+t16)*t26
      dgu(1,1,1) = (t3+t6+t9+t12)*t22-t27*t33
      dgu(1,1,2) = (t37+t39+t41+t43)*t22-t27*t50
      dgu(1,1,3) = (t54+t56+t58+t60)*t22-t27*t67
      dgu(1,2,1) = 2*t80
      dgu(1,2,2) = 2*t89
      dgu(1,2,3) = 2*t98
      dgu(1,3,1) = 2*t107
      dgu(1,3,2) = 2*t114
      dgu(1,3,3) = 2*t121
      dgu(2,1,1) = 2*t80
      dgu(2,1,2) = 2*t89
      dgu(2,1,3) = 2*t98
      dgu(2,2,1) = (t3+t6+t29+t32)*t22-t125*t33
      dgu(2,2,2) = (t37+t39+t47+t49)*t22-t125*t50
      dgu(2,2,3) = (t54+t56+t64+t66)*t22-t125*t67
      dgu(2,3,1) = 2*t142
      dgu(2,3,2) = 2*t149
      dgu(2,3,3) = 2*t156
      dgu(3,1,1) = 2*t107
      dgu(3,1,2) = 2*t114
      dgu(3,1,3) = 2*t121
      dgu(3,2,1) = 2*t142
      dgu(3,2,2) = 2*t149
      dgu(3,2,3) = 2*t156
      dgu(3,3,1) = (t9+t12+t29+t32)*t22-t160*t33
      dgu(3,3,2) = (t41+t43+t47+t49)*t22-t160*t50
      dgu(3,3,3) = (t58+t60+t64+t66)*t22-t160*t67
      dxguxx=dgu(1,1,1)
      dyguxx=dgu(1,1,2)
      dzguxx=dgu(1,1,3)
      dxguxy=dgu(1,2,1)
      dyguxy=dgu(1,2,2)
      dzguxy=dgu(1,2,3)
      dxguxz=dgu(1,3,1)
      dyguxz=dgu(1,3,2)
      dzguxz=dgu(1,3,3)
      dxguyy=dgu(2,2,1)
      dyguyy=dgu(2,2,2)
      dzguyy=dgu(2,2,3)
      dxguyz=dgu(2,3,1)
      dyguyz=dgu(2,3,2)
      dzguyz=dgu(2,3,3)
      dxguzz=dgu(3,3,1)
      dyguzz=dgu(3,3,2)
      dzguzz=dgu(3,3,3)
      betax = guxx*betadx+guxy*betady+guxz*betadz
      betay = guxy*betadx+guyy*betady+guyz*betadz
      betaz = guxz*betadx+guyz*betady+guzz*betadz
      dxbetax = dxguxx*betadx+guxx*dxbetadx+dxguxy*betady+guxy*dxbetady+&
     &dxguxz*betadz+guxz*dxbetadz
      dxbetay = dxguxy*betadx+guxy*dxbetadx+dxguyy*betady+guyy*dxbetady+&
     &dxguyz*betadz+guyz*dxbetadz
      dxbetaz = dxguxz*betadx+guxz*dxbetadx+dxguyz*betady+guyz*dxbetady+&
     &dxguzz*betadz+guzz*dxbetadz
      dybetax = dyguxx*betadx+guxx*dybetadx+dyguxy*betady+guxy*dybetady+&
     &dyguxz*betadz+guxz*dybetadz
      dybetay = dyguxy*betadx+guxy*dybetadx+dyguyy*betady+guyy*dybetady+&
     &dyguyz*betadz+guyz*dybetadz
      dybetaz = dyguxz*betadx+guxz*dybetadx+dyguyz*betady+guyz*dybetady+&
     &dyguzz*betadz+guzz*dybetadz
      dzbetax = dzguxx*betadx+guxx*dzbetadx+dzguxy*betady+guxy*dzbetady+&
     &dzguxz*betadz+guxz*dzbetadz
      dzbetay = dzguxy*betadx+guxy*dzbetadx+dzguyy*betady+guyy*dzbetady+&
     &dzguyz*betadz+guyz*dzbetadz
      dzbetaz = dzguxz*betadx+guxz*dzbetadx+dzguyz*betady+guyz*dzbetady+&
     &dzguzz*betadz+guzz*dzbetadz

      betau(1) = betax
      betau(2) = betay
      betau(3) = betaz
      dbetau(1,1) = dxbetax
      dbetau(1,2) = dybetax
      dbetau(1,3) = dzbetax
      dbetau(2,1) = dxbetay
      dbetau(2,2) = dybetay
      dbetau(2,3) = dzbetay
      dbetau(3,1) = dxbetaz
      dbetau(3,2) = dybetaz
      dbetau(3,3) = dzbetaz
     
      call calB(betau, , dbetau, Bu, dBu, xx, par)

      Bx = Bu(1)
      By = Bu(2)
      Bz = Bu(3)

      dxBx = dBu(1,1)
      dyBx = dBu(1,2)
      dzBx = dBu(1,3)
      dxBy = dBu(2,1)
      dyBy = dBu(2,2)
      dzBy = dBu(2,3)
      dxBz = dBu(3,1)
      dyBz = dBu(3,2)
      dzBz = dBu(3,3)

      c111 = 0.5D0*guxx*dxgxx+guxy*(0.1D1*dxgxy-0.5D0*dygxx)+guxz*(0.1D1&
     &*dxgxz-0.5D0*dzgxx)
      c112 = 0.5D0*guxx*dygxx+0.5D0*guxy*dxgyy+guxz*(0.5D0*dygxz+0.5D0*d&
     &xgyz-0.5D0*dzgxy)
      c113 = 0.5D0*guxx*dzgxx+guxy*(0.5D0*dzgxy+0.5D0*dxgyz-0.5D0*dygxz)&
     &+0.5D0*guxz*dxgzz
      c122 = guxx*(0.1D1*dygxy-0.5D0*dxgyy)+0.5D0*guxy*dygyy+guxz*(0.1D1&
     &*dygyz-0.5D0*dzgyy)
      c123 = guxx*(0.5D0*dzgxy+0.5D0*dygxz-0.5D0*dxgyz)+0.5D0*guxy*dzgyy&
     &+0.5D0*guxz*dygzz
      c133 = guxx*(0.1D1*dzgxz-0.5D0*dxgzz)+guxy*(0.1D1*dzgyz-0.5D0*dygz&
     &z)+0.5D0*guxz*dzgzz
      c211 = 0.5D0*guxy*dxgxx+guyy*(0.1D1*dxgxy-0.5D0*dygxx)+guyz*(0.1D1&
     &*dxgxz-0.5D0*dzgxx)
      c212 = 0.5D0*guxy*dygxx+0.5D0*guyy*dxgyy+guyz*(0.5D0*dygxz+0.5D0*d&
     &xgyz-0.5D0*dzgxy)
      c213 = 0.5D0*guxy*dzgxx+guyy*(0.5D0*dzgxy+0.5D0*dxgyz-0.5D0*dygxz)&
     &+0.5D0*guyz*dxgzz
      c222 = guxy*(0.1D1*dygxy-0.5D0*dxgyy)+0.5D0*guyy*dygyy+guyz*(0.1D1&
     &*dygyz-0.5D0*dzgyy)
      c223 = guxy*(0.5D0*dzgxy+0.5D0*dygxz-0.5D0*dxgyz)+0.5D0*guyy*dzgyy&
     &+0.5D0*guyz*dygzz
      c233 = guxy*(0.1D1*dzgxz-0.5D0*dxgzz)+guyy*(0.1D1*dzgyz-0.5D0*dygz&
     &z)+0.5D0*guyz*dzgzz
      c311 = 0.5D0*guxz*dxgxx+guyz*(0.1D1*dxgxy-0.5D0*dygxx)+guzz*(0.1D1&
     &*dxgxz-0.5D0*dzgxx)
      c312 = 0.5D0*guxz*dygxx+0.5D0*guyz*dxgyy+guzz*(0.5D0*dygxz+0.5D0*d&
     &xgyz-0.5D0*dzgxy)
      c313 = 0.5D0*guxz*dzgxx+guyz*(0.5D0*dzgxy+0.5D0*dxgyz-0.5D0*dygxz)&
     &+0.5D0*guzz*dxgzz
      c322 = guxz*(0.1D1*dygxy-0.5D0*dxgyy)+0.5D0*guyz*dygyy+guzz*(0.1D1&
     &*dygyz-0.5D0*dzgyy)
      c323 = guxz*(0.5D0*dzgxy+0.5D0*dygxz-0.5D0*dxgyz)+0.5D0*guyz*dzgyy&
     &+0.5D0*guzz*dygzz
      c333 = guxz*(0.1D1*dzgxz-0.5D0*dxgzz)+guyz*(0.1D1*dzgyz-0.5D0*dygz&
     &z)+0.5D0*guzz*dzgzz
      t1 = 1/alpha
      s1 = guxx*t1*(-dtgxx+betax*dxgxx+betay*dygxx+betaz*dzgxx+2*gxx*dxb&
     &etax+2*gxy*dxbetay+2*gxz*dxbetaz)/2+guxy*t1*(-dtgxy+betax*dxgxy+be&
     &tay*dygxy+betaz*dzgxy+gxy*dxbetax+gyy*dxbetay+gyz*dxbetaz+gxx*dybe&
     &tax+gxy*dybetay+gxz*dybetaz)+guxz*t1*(-dtgxz+betax*dxgxz+betay*dyg&
     &xz+betaz*dzgxz+gxz*dxbetax+gyz*dxbetay+gzz*dxbetaz+gxx*dzbetax+gxy&
     &*dzbetay+gxz*dzbetaz)
      trK = s1+guyy*t1*(-dtgyy+betax*dxgyy+betay*dygyy+betaz*dzgyy+2*gxy&
     &*dybetax+2*gyy*dybetay+2*gyz*dybetaz)/2+guyz*t1*(-dtgyz+betax*dxgy&
     &z+betay*dygyz+betaz*dzgyz+gxz*dybetax+gyz*dybetay+gzz*dybetaz+gxy*&
     &dzbetax+gyy*dzbetay+gyz*dzbetaz)+guzz*t1*(-dtgzz+betax*dxgzz+betay&
     &*dygzz+betaz*dzgzz+2*gxz*dzbetax+2*gyz*dzbetay+2*gzz*dzbetaz)/2
      t87 = guxx*c111*P1+guxx*c211*P2+guxx*c311*P3+2*guxy*c112*P1+2*guxy&
     &*c212*P2+2*guxy*c312*P3+2*guxz*c113*P1+2*guxz*c213*P2+2*guxz*c313*&
     &P3+guyy*c122*P1+guyy*c222*P2+guyy*c322*P3+2*guyz*c123*P1+2*guyz*c2&
     &23*P2+2*guyz*c323*P3+guzz*c133*P1+guzz*c233*P2+guzz*c333*P3
      t91 = guxx*(dxalpha*P1+alpha*dxP1)+guxy*(dxalpha*P2+alpha*dxP2)+gu&
     &xz*(dxalpha*P3+alpha*dxP3)+guxy*(dyalpha*P1+alpha*dyP1)+guyy*(dyal&
     &pha*P2+alpha*dyP2)+guyz*(dyalpha*P3+alpha*dyP3)+guxz*(dzalpha*P1+a&
     &lpha*dzP1)+guyz*(dzalpha*P2+alpha*dzP2)+guzz*(dzalpha*P3+alpha*dzP&
     &3)-alpha*t87+alpha*trK*PI
      t93 = 1/alpha
      t94 = betax-Bx
      t100 = betay-By
      t104 = betaz-Bz
      t145 = alpha**2
      t146 = 1/t145
      t173 = dxbetax*t93-betax*t146*dxalpha+dybetay*t93-betay*t146*dyalp&
     &ha+dzbetaz*t93-betaz*t146*dzalpha+c111*betax*t93+c112*betay*t93+c1&
     &13*betaz*t93+c212*betax*t93+c222*betay*t93+c223*betaz*t93+c313*bet&
     &ax*t93+c323*betay*t93+c333*betaz*t93
      s1 = -alpha*Vsf-t93*(t94*((dxbetax-dxBx)*P1+t94*dxP1+(dxbetay-dxBy&
     &)*P2+t100*dxP2+(dxbetaz-dxBz)*P3+t104*dxP3)+t100*((dybetax-dyBx)*P&
     &1+t94*dyP1+(dybetay-dyBy)*P2+t100*dyP2+(dybetaz-dyBz)*P3+t104*dyP3&
     &)+t104*((dzbetax-dzBx)*P1+t94*dzP1+(dzbetay-dzBy)*P2+t100*dzP2+(dz&
     &betaz-dzBz)*P3+t104*dzP3))+2*t94*dxPI+2*t100*dyPI+2*t104*dzPI
      t220 = s1+t93*PI*(t94*dxalpha+t100*dyalpha+t104*dzalpha)-t173*(t94&
     &*P1+t100*P2+t104*P3)+t93*((betax*dxBx-Bx*dxbetax)*P1+(betax*dxBy-B&
     &x*dxbetay)*P2+(betax*dxBz-Bx*dxbetaz)*P3+(betay*dyBx-By*dybetax)*P&
     &1+(betay*dyBy-By*dybetay)*P2+(betay*dyBz-By*dybetaz)*P3+(betaz*dzB&
     &x-Bz*dzbetax)*P1+(betaz*dzBy-Bz*dzbetay)*P2+(betaz*dzBz-Bz*dzbetaz&
     &)*P3)+Bx*dxPI+By*dyPI+Bz*dzPI

      RHS(H_PHI) = alpha*PI+Bx*dxPHI+By*dyPHI+Bz*dzPHI
      RHS(H_D1PHI) = dxalpha*PI+alpha*dxPI+Bx*dxP1+By*dyP1+Bz*dzP1&
     &+P1*dxBx+P2*dxBy+P3*dxBz
      RHS(H_D2PHI) = dyalpha*PI+alpha*dyPI+Bx*dxP2+By*dyP2+Bz*dzP2&
     &+P1*dyBx+P2*dyBy+P3*dyBz
      RHS(H_D3PHI) = dzalpha*PI+alpha*dzPI+Bx*dxP3+By*dyP3+Bz*dzP3&
     &+P1*dzBx+P2*dzBy+P3*dzBz
      RHS(H_D4PHI) = t91+t220

      do i = 1, NU
        if (.not.(RHS(i) .eq. 1.0d0*RHS(i))) then
          write(0,*)'must be a NaN: ',RHS(i)
          stop
        end if
      end do

      return
      end subroutine

end module
