cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  comp_amr_error:                                                           cc
cc                   Needs to set the field gr_error to have                  cc
cc                   the appropriate error function....                       cc
cc                   therefore needs to be non-negative function              cc
cc                   of the fields.                                           cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine comp_amr_error( gridnum )
      implicit    none
      integer     gridnum
      include    'grid.inc'
      include    'glob.inc'
      include    'param.inc'
      integer     nx, ny, nz, numfields
      real*8      hg,  tmp, tmp_maxchi
      integer               tmp_maxchi_i, tmp_maxchi_j,tmp_maxchi_k
      logical     ltrace
      parameter ( ltrace = .false. )
      integer    i,j,k,index, l
      real*8     grad_x_psi, grad_y_psi, grad_z_psi

      call load_pointers(gridnum)

      if (ltrace) write(*,*) 'comp_amr_error: On grid: ',gridnum

      nx   = gr_nx(gridnum)
      ny   = gr_ny(gridnum)
      nz   = gr_nz(gridnum)
      hg   = gr_h(gridnum)


      if ( shadow .eq. 0 ) then
      !
      ! GENERIC computation:
      !
      if (.false.) then
         !
         ! Loop over all integral fields and add them up in quadrature
         !
         if(ltrace)write(*,*)'comp_amr_error: Looping over INTEGRALflds'
         numfields = 0
         do l = 1, num_gfuncs - 1
            if (gfunc_type(l) .eq. GFUNC_INTEGRAL ) then
               numfields = numfields + 1
               if(ltrace)write(*,*)'comp_amr_error:field:',gfunc_name(l)
               do k = 1, nz
               do j = 1, ny
               do i = 1, nx
                  index = (k-1)*ny*nx + (j-1)*nx + i
                  q(gr_error + index - 1) =
     *                      q(gr_error+index-1)
     *                    + q(gfunc_pointer(l,gridnum)+index-1)**2
               end do
               end do
               end do
            end if
         end do
         !
         if (ltrace) write(*,*) 'comp_amr_error: numfields=',numfields
         if (numfields .gt. 0) then
            tmp = hg**2 / numfields
            do k = 1, nz
            do j = 1, ny
            do i = 1, nx
                     index = (k-1)*ny*nx + (j-1)*nx + i
                     q(gr_error + index - 1) = 
     *                           q(gr_error+index-1) * tmp
            end do
            end do
            end do
         end if
      end if

      end if ! END : if shadow .eq. 0
       


      return
      end    ! END: comp_amr_error


cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_surface_comp:                                                        cc
cc                   Takes as input the coordinate of a point located         cc
cc                   on the grid. It needs to compute the various             cc
cc                   surface functions at that point.                         cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_surface_comp(gi, x,y,z, itheta, jphi)
      implicit     none
      integer      gi, itheta, jphi
      real(kind=8) x,y,z
      include     'grid.inc'
      include     'surfaces.inc'
      integer      nx, ny, nz
      integer      i,j,k,index, l
      integer      ilow, jlow, klow
      real(kind=8) hg, phi, pi, fracx, fracy, fracz
      real(kind=8) interp
      external     interp
      !
      logical     ltrace
      parameter ( ltrace = .false. )

      return

      call load_pointers(gi)

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)

      ! Index of lower left point of interpolation cell:
      ilow  = INT( (x - gr_minx(gi)) / hg ) + 1
      jlow  = INT( (y - gr_miny(gi)) / hg ) + 1
      klow  = INT( (z - gr_minz(gi)) / hg ) + 1

      ! Fractional distance point is from lower left point:
      fracx = ( x - q( gr_x(gi) + (ilow-1) ) ) / hg
      fracy = ( y - q( gr_y(gi) + (jlow-1) ) ) / hg
      fracz = ( z - q( gr_z(gi) + (klow-1) ) ) / hg

      if (ilow .lt. 1) then
         write(*,*) 'grid_surface_comp: x too small ',ilow
         call my_exit('grid_surface_comp: X too small')
      end if
      if (fracx .lt. 0 .or. fracy.lt.0 .or. fracz.lt.0) then
         write(*,*) 'grid_surface_comp: fraction cannot be negative'
         write(*,*) 'grid_surface_comp: fracx/y/z =',fracx,fracy,fracz
         call my_exit('grid_surface_comp: fraction cannot be negative')
      end if

!     phi = interp(  q(gr_phi),
!    *               fracx, fracy, fracz,
!    *               ilow,jlow,klow,
!    *               nx,ny,nz)
!     pi  = interp(  q(gr_pi),
!    *               fracx, fracy, fracz,
!    *               ilow,jlow,klow,
!    *               nx,ny,nz)


!     index = (jphi-1)*asf_shape(1) + (itheta-1)
!     q(asf_q+index) = phi
!     q(asf_g+index) = pi

      if (ltrace) then
         write(*,*) 'grid_surface_comp: ********'
         write(*,*) 'grid_surface_comp: gi,hg     =',gi,hg
         write(*,9) 'grid_surface_comp: x,y,z     =',x,y,z
         write(*,8) 'grid_surface_comp: nx/y/z    =',nx,ny,nz
         write(*,8) 'grid_surface_comp: i/j/klow  =',ilow,jlow,klow
         write(*,9) 'grid_surface_comp: fracx/y/z =',fracx,fracy,fracz
         write(*,9) 'grid_surface_comp: xmin/ymin/zmin =',
     *                gr_minx(gi),gr_miny(gi),gr_minz(gi)
      end if

 8    format(A,5I10)
 9    format(A,5F10.4)

      return
      end    ! END: grid_surface_comp

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_ahfind:                                                              cc
cc                                                                            cc
cc         Project specific routine to find where to excise and hence         cc
cc         should be located in comp_amr_error.f.                             cc
cc         This routines requires centers for up to masks to be specified.    cc
cc         All this is subject to change, but for now this routine is         cc
cc         responsible for determining on this grid whether the quantities    cc
cc             local_left(1..2,1..3)                                          cc
cc             local_right(1..2,1..3)                                         cc
cc         need to be adjusted. These are local to this processor, and        cc
cc         global bounds will be found in level_bhmask_local().               cc
cc                                                                            cc
cc         To see the data structures for this routine, look in:              cc
cc                  had/include/mask.inc                                      cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_ahfind(gi)
      implicit none
      integer  gi
      include 'mpif.h'
      include 'mpi_stuff.inc'
      include 'grid.inc'
      include 'mask.inc'
      include 'param.inc'
      integer  nx,ny,nz, i,j,k, kk
      real(kind=8) my_left(3,max_num_masks),
     *             my_right(3,max_num_masks)
      logical     bh_found

      logical     ltrace
      parameter ( ltrace  = .false.)
      logical     ltrace2
      parameter ( ltrace2 = .false.)

      call load_pointers(gi)

      nx = gr_nx(gi)
      ny = gr_ny(gi)
      nz = gr_nz(gi)

      num_masks = 0

      !
      ! Do the real computation for this grid:
      !
      do kk = 1, max_num_masks
        if (ltrace2) then
           write(*,99) myid, 'grid_ahfind: Center:',kk
           do i = 1, 3
              write(*,98) myid, '    center coord',mask_center(i,kk)
           end do
        end if
        !
        bh_found    = .false.

        if ( bh_found ) then
            bh_true(kk) = .true.
            num_masks = num_masks + 1
            if(ltrace)write(*,99)myid, 'grid_ahfind: BH found: ',kk
            do i = 1, 3
               if (ltrace2)write(*,98)myid, ' l/right:',
     *                              my_left(i,kk),my_right(i,kk)
               if (my_left(i,kk)  .lt. local_left(i,kk)  )
     *             local_left(i,kk)  = my_left(i,kk)
               if (my_right(i,kk) .gt. local_right(i,kk) )
     *             local_right(i,kk) = my_right(i,kk)
               if(ltrace)write(*,98)myid,'      local_left/right(i,k)=',
     *            local_left(i,kk), local_right(i,kk)
            end do
        end if
         !
      end do

      if (ltrace2) then
         write(*,99) myid, 'grid_ahfind: gi     = ',gi
         write(*,99) myid, 'grid_ahfind: nx/y/z = ',nx,ny,nz
         write(*,99) myid, 'grid_ahfind: Done.'
      end if

  98  format('[',I3,'] ',A,3F18.6)
  99  format('[',I3,'] ',A,3I5)
      return
      end       ! END: grid_ahfind
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  compute_boundary_bh:                                                      cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine compute_boundary_bh(alpha,betax,betay,betaz,
     *           g11, g12, g13, g22, g23, g33,
     *           k11, k12, k13, k22, k23, k33,
     *           d1g11, d1g12, d1g13, d1g22, d1g23, d1g33,
     *           d2g11, d2g12, d2g13, d2g22, d2g23, d2g33,
     *           d3g11, d3g12, d3g13, d3g22, d3g23, d3g33, 
     *           rad_exp,rad_speed,
     *           xx,   yy,   zz,
     *           bh_center, bh_radius, bh_min, bh_max, bh_true,
     *           nx, ny, nz)
      implicit   none
      integer        nx,ny,nz
      real(kind=8)   alpha(nx,ny,nz), betax(nx,ny,nz), betay(nx,ny,nz),
     *               betaz(nx,ny,nz),
     *               g11(nx,ny,nz), g12(nx,ny,nz), g13(nx,ny,nz),
     *               g22(nx,ny,nz), g23(nx,ny,nz), g33(nx,ny,nz),
     *               d1g11(nx,ny,nz), d1g12(nx,ny,nz), d1g13(nx,ny,nz),
     *               d1g22(nx,ny,nz), d1g23(nx,ny,nz), d1g33(nx,ny,nz),
     *               d2g11(nx,ny,nz), d2g12(nx,ny,nz), d2g13(nx,ny,nz),
     *               d2g22(nx,ny,nz), d2g23(nx,ny,nz), d2g33(nx,ny,nz),
     *               d3g11(nx,ny,nz), d3g12(nx,ny,nz), d3g13(nx,ny,nz),
     *               d3g22(nx,ny,nz), d3g23(nx,ny,nz), d3g33(nx,ny,nz),
     *               k11(nx,ny,nz), k12(nx,ny,nz), k13(nx,ny,nz),
     *               k22(nx,ny,nz), k23(nx,ny,nz), k33(nx,ny,nz),
     *               rad_exp(nx,ny,nz),rad_speed(nx,ny,nz),
     *               xx(nx),        yy(ny),        zz(nz)
      real(kind=8)   bh_center(3), bh_radius
      real(kind=8)   bh_min(3),     bh_max(3)
      logical        bh_true
      ! Local vars:
      integer        i,j,k
      real(kind=8)   x,y,z,r,h11,h12,h13,h22,h23,h33,
     *               huu11,huu12,huu13,huu22,huu23,huu33,
     *               deth, b1, b2, b3, alp, alp2,
     *               eK11,eK12,eK13,eK22,eK23,eK33,etrK,     
     *               nr1, nr2, nr3, nru1, nru2, nru3, nrnorm,
     *               Dddu111,Dddu112,Dddu113,Dddu121,Dddu122,Dddu123,
     *               Dddu131,Dddu132,Dddu133,    
     *               Dddu211,Dddu212,Dddu213,Dddu221,Dddu222,Dddu223,
     *               Dddu231,Dddu232,Dddu233,    
     *               Dddu311,Dddu312,Dddu313,Dddu321,Dddu322,Dddu323,
     *               Dddu331,Dddu332,Dddu333,    
     *               Dudd111,Dudd112,Dudd113,Dudd122,Dudd123,Dudd133,
     *               Dudd211,Dudd212,Dudd213,Dudd222,Dudd223,Dudd233,
     *               Dudd311,Dudd312,Dudd313,Dudd322,Dudd323,Dudd333,     
     *               G111,G112,G113,G122,G123,G133,
     *               G211,G212,G213,G222,G223,G233,
     *               G311,G312,G313,G322,G323,G333,
     *               Gndd11,Gndd12,Gndd13,Gndd22,Gndd23,Gndd33,
     *               Dndd11,Dndd12,Dndd13,Dndd22,Dndd23,Dndd33,
     *               D111,D122,D211,D222,D311,D322,
     *               D112,D113,D123,D133,D212,D213,
     *               D312,D313,D323,D333,D233,D223

      logical     ltrace
      parameter ( ltrace = .false.)
      

      bh_true = .false.
      !
      bh_min(1) = xx(nx)
      bh_min(2) = yy(ny)
      bh_min(3) = zz(nz)
      !
      bh_max(1) = xx(1)
      bh_max(2) = yy(1)
      bh_max(3) = zz(1)

      do k = 1, nz
         z = zz(k)
         !
         do j = 1, ny
            y = yy(j)
            !
            do i = 1, nx
               x = xx(i)
               !
               !       compute the 3+1 quantities
               !
               h11 = g11(i,j,k)
               h12 = g12(i,j,k)
               h13 = g13(i,j,k)
               h22 = g22(i,j,k)
               h23 = g23(i,j,k)
               h33 = g33(i,j,k)
               !
               huu11 = -h23**2 + h22*h33
               huu12 = h13*h23 - h12*h33
               huu13 = -(h13*h22) + h12*h23
               huu22 = -h13**2 + h11*h33
               huu23 = h12*h13 - h11*h23
               huu33 = -h12**2 + h11*h22
               deth  = h11*huu11 + h12*huu12 + h13*huu13
               huu11 = huu11/deth
               huu12 = huu12/deth
               huu13 = huu13/deth
               huu22 = huu22/deth
               huu23 = huu23/deth
               huu33 = huu33/deth
               !
               ! compute the speed in the radial direction
               !
               b1 = betax(i,j,k)
               b2 = betay(i,j,k)
               b3 = betaz(i,j,k)
               alp = alpha(i,j,k)

               nr1    = (x - bh_center(1)) + 1d-12
               nr2    = (y - bh_center(2)) + 1d-12
               nr3    = (z - bh_center(3)) + 1d-12
	       
               r      = dsqrt(nr1**2 + nr2**2 + nr3**2)

               if ( r .gt. 0.0005 ) then  
	       
               nrnorm = dsqrt(   huu11*nr1*nr1 + huu22*nr2*nr2
     *                         + huu33*nr3*nr3
     *                    +2.0d0*(huu12*nr1*nr2 + huu13*nr1*nr3
     *                            + huu23*nr2*nr3))
               nr1 = nr1/nrnorm
               nr2 = nr2/nrnorm
               nr3 = nr3/nrnorm      	       	       
	       
               !
               rad_speed(i,j,k) = - (nr1*b1 + nr2*b2 + nr3*b3) + alp
               !
               !
               ! compute the expansion in the radial direction
               !
	       
	       nru1 = huu11*nr1 + huu12*nr2 + huu13*nr3
               nru2 = huu12*nr1 + huu22*nr2 + huu23*nr3
               nru3 = huu13*nr1 + huu23*nr2 + huu33*nr3

	       D111 = d1g11(i,j,k); D112 = d1g12(i,j,k); D113 = d1g13(i,j,k)
	       D122 = d1g22(i,j,k); D123 = d1g23(i,j,k); D133 = d1g33(i,j,k)
	       D211 = d2g11(i,j,k); D212 = d2g12(i,j,k); D213 = d2g13(i,j,k)
	       D222 = d2g22(i,j,k); D223 = d2g23(i,j,k); D233 = d2g33(i,j,k)
	       D311 = d3g11(i,j,k); D312 = d3g12(i,j,k); D313 = d3g13(i,j,k)
	       D322 = d3g22(i,j,k); D323 = d3g23(i,j,k); D333 = d3g33(i,j,k)

	       	       
               eK11 = k11(i,j,k)
               eK12 = k12(i,j,k)
               eK13 = k13(i,j,k)
               eK22 = k22(i,j,k)
               eK23 = k23(i,j,k)
               eK33 = k33(i,j,k)
	       
               etrK = huu11*eK11 + huu12*eK12 + huu13*eK13
     *              + huu12*eK12 + huu22*eK22 + huu23*eK23
     *              + huu13*eK13 + huu23*eK23 + huu33*eK33
	       
               Dddu111 = huu11*D111 + huu12*D112 + huu13*D113
               Dddu112 = huu12*D111 + huu22*D112 + huu23*D113
               Dddu113 = huu13*D111 + huu23*D112 + huu33*D113
               Dddu121 = huu11*D112 + huu12*D122 + huu13*D123
               Dddu122 = huu12*D112 + huu22*D122 + huu23*D123
               Dddu123 = huu13*D112 + huu23*D122 + huu33*D123
               Dddu131 = huu11*D113 + huu12*D123 + huu13*D133
               Dddu132 = huu12*D113 + huu22*D123 + huu23*D133
               Dddu133 = huu13*D113 + huu23*D123 + huu33*D133
               Dddu211 = huu11*D211 + huu12*D212 + huu13*D213
               Dddu212 = huu12*D211 + huu22*D212 + huu23*D213
               Dddu213 = huu13*D211 + huu23*D212 + huu33*D213
               Dddu221 = huu11*D212 + huu12*D222 + huu13*D223
               Dddu222 = huu12*D212 + huu22*D222 + huu23*D223
               Dddu223 = huu13*D212 + huu23*D222 + huu33*D223
               Dddu231 = huu11*D213 + huu12*D223 + huu13*D233
               Dddu232 = huu12*D213 + huu22*D223 + huu23*D233
               Dddu233 = huu13*D213 + huu23*D223 + huu33*D233
               Dddu311 = huu11*D311 + huu12*D312 + huu13*D313
               Dddu312 = huu12*D311 + huu22*D312 + huu23*D313
               Dddu313 = huu13*D311 + huu23*D312 + huu33*D313
               Dddu321 = huu11*D312 + huu12*D322 + huu13*D323
               Dddu322 = huu12*D312 + huu22*D322 + huu23*D323
               Dddu323 = huu13*D312 + huu23*D322 + huu33*D323
               Dddu331 = huu11*D313 + huu12*D323 + huu13*D333
               Dddu332 = huu12*D313 + huu22*D323 + huu23*D333
               Dddu333 = huu13*D313 + huu23*D323 + huu33*D333

               Dudd111 = huu11*D111 + huu12*D211 + huu13*D311
               Dudd112 = huu11*D112 + huu12*D212 + huu13*D312
               Dudd113 = huu11*D113 + huu12*D213 + huu13*D313
               Dudd122 = huu11*D122 + huu12*D222 + huu13*D322
               Dudd123 = huu11*D123 + huu12*D223 + huu13*D323
               Dudd133 = huu11*D133 + huu12*D233 + huu13*D333
               Dudd211 = huu12*D111 + huu22*D211 + huu23*D311
               Dudd212 = huu12*D112 + huu22*D212 + huu23*D312
               Dudd213 = huu12*D113 + huu22*D213 + huu23*D313
               Dudd222 = huu12*D122 + huu22*D222 + huu23*D322
               Dudd223 = huu12*D123 + huu22*D223 + huu23*D323
               Dudd233 = huu12*D133 + huu22*D233 + huu23*D333
               Dudd311 = huu13*D111 + huu23*D211 + huu33*D311
               Dudd312 = huu13*D112 + huu23*D212 + huu33*D312
               Dudd313 = huu13*D113 + huu23*D213 + huu33*D313
               Dudd322 = huu13*D122 + huu23*D222 + huu33*D322
               Dudd323 = huu13*D123 + huu23*D223 + huu33*D323
               Dudd333 = huu13*D133 + huu23*D233 + huu33*D333
	       	       
              !with the first index up                                             
               G111 = Dddu111 - Dudd111/2.
               G112 = (Dddu121 + Dddu211 - Dudd112)/2.
               G113 = (Dddu131 + Dddu311 - Dudd113)/2.
               G122 = Dddu221 - Dudd122/2.
               G123 = (Dddu231 + Dddu321 - Dudd123)/2.
               G133 = Dddu331 - Dudd133/2.
               G211 = Dddu112 - Dudd211/2.
               G212 = (Dddu122 + Dddu212 - Dudd212)/2.
               G213 = (Dddu132 + Dddu312 - Dudd213)/2.
               G222 = Dddu222 - Dudd222/2.
               G223 = (Dddu232 + Dddu322 - Dudd223)/2.
               G233 = Dddu332 - Dudd233/2.
               G311 = Dddu113 - Dudd311/2.
               G312 = (Dddu123 + Dddu213 - Dudd312)/2.
               G313 = (Dddu133 + Dddu313 - Dudd313)/2.
               G322 = Dddu223 - Dudd322/2.
               G323 = (Dddu233 + Dddu323 - Dudd323)/2.
               G333 = Dddu333 - Dudd333/2.

               Gndd11 = nr1*G111 + nr2*G211 + nr3*G311
               Gndd12 = nr1*G112 + nr2*G212 + nr3*G312
               Gndd13 = nr1*G113 + nr2*G213 + nr3*G313
               Gndd22 = nr1*G122 + nr2*G222 + nr3*G322
               Gndd23 = nr1*G123 + nr2*G223 + nr3*G323
               Gndd33 = nr1*G133 + nr2*G233 + nr3*G333

               Dndd11 = nru1*D111 + nru2*D211 + nru3*D311
               Dndd12 = nru1*D112 + nru2*D212 + nru3*D312
               Dndd13 = nru1*D113 + nru2*D213 + nru3*D313
               Dndd22 = nru1*D122 + nru2*D222 + nru3*D322
               Dndd23 = nru1*D123 + nru2*D223 + nru3*D323
               Dndd33 = nru1*D133 + nru2*D233 + nru3*D333
	                     
               rad_exp(i,j,k) = - etrK + ( (huu11 - nru1*nru1) 
     *         + (huu22 - nru2*nru2) + (huu33 - nru3*nru3) )/nrnorm 
     *         - ( huu11*Gndd11 + huu22*Gndd22 + huu33*Gndd33 
     *         + 2.0*(huu12*Gndd12 + huu13*Gndd13 + huu23*Gndd23) ) 
     *         + nru1*nru1*(eK11  + 0.5*Dndd11) 
     *         + nru2*nru2*(eK22  + 0.5*Dndd22)  
     *         + nru3*nru3*(eK33  + 0.5*Dndd33)        
     *         + 2.0*nru1*nru2*(eK12  + 0.5*Dndd12)  
     *         + 2.0*nru1*nru3*(eK13  + 0.5*Dndd13)  
     *         + 2.0*nru2*nru3*(eK23  + 0.5*Dndd23)         
                    	       
	      if ((rad_speed(i,j,k) .lt. 0) .AND. 
     *            (rad_exp(i,j,k) .lt. 0) .AND.
     *            (r .lt. 2.5*bh_radius))  then
                  bh_true = .true.
                  if (x .lt. bh_min(1)) bh_min(1) = x
                  if (y .lt. bh_min(2)) bh_min(2) = y
                  if (z .lt. bh_min(3)) bh_min(3) = z	    
                  if (x .gt. bh_max(1)) bh_max(1) = x
                  if (y .gt. bh_max(2)) bh_max(2) = y
                  if (z .gt. bh_max(3)) bh_max(3) = z		  
              end if  
              else
                rad_exp(i,j,k) = 0.0;
                rad_speed(i,j,k) = 0.0;
              end if
            end do  
         end do  
      end do  

      if (ltrace) then
         write(*,*) 'compute_boundary_bh: bh_true = ',bh_true
         if (bh_true) then
            do i = 1, 3
               write(*,*)'compute_boundary_bh: bh_center: ',bh_center(i)
            end do
            do i = 1, 3
               write(*,*)'compute_boundary_bh: min/max: ',
     *                                     bh_min(i),bh_max(i)
            end do
         end if
      end if

      return
      end subroutine compute_boundary_bh
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_asurface_comp:                                                        cc
cc                   Takes as input the coordinate of a point located         cc
cc                   on the grid. It needs to compute the various             cc
cc                   surface functions at that point.                         cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_asurface_comp(gi, x,y,z, itheta, jphi)
      implicit     none
      integer      gi, itheta, jphi
      real(kind=8) x,y,z
      include     'grid.inc'
      include     'surfaces.inc'
      integer      nx, ny, nz
      integer      i,j,k,index, l
      integer      ilow, jlow, klow
      real(kind=8) hg, rpsi4, ipsi4, fracx, fracy, fracz
      real(kind=8) interp
      external     interp
      !
      logical     ltrace
      parameter ( ltrace = .false. )

      return

      call load_pointers(gi)

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)

      ! Index of lower left point of interpolation cell:
      ilow  = INT( (x - gr_minx(gi)) / hg ) + 1
      jlow  = INT( (y - gr_miny(gi)) / hg ) + 1
      klow  = INT( (z - gr_minz(gi)) / hg ) + 1

      ! Fractional distance point is from lower left point:
      fracx = ( x - q( gr_x(gi) + (ilow-1) ) ) / hg
      fracy = ( y - q( gr_y(gi) + (jlow-1) ) ) / hg
      fracz = ( z - q( gr_z(gi) + (klow-1) ) ) / hg

      if (ilow .lt. 1) then
         write(*,*) 'grid_surface_comp: x too small ',ilow
         call my_exit('grid_surface_comp: X too small')
      end if
      if (fracx.lt.-0.01 .or.fracy.lt.-0.01.or.fracz.lt.-0.0001) then
         write(*,*) 'grid_surface_comp: fraction cannot be negative'
         write(*,*) 'grid_surface_comp: fracx/y/z =',fracx,fracy,fracz
         call my_exit('grid_surface_comp: fraction cannot be negative')
      end if

!      rpsi4 = interp(  q(gr_psi4R),
!     *               fracx, fracy, fracz,
!     *               ilow,jlow,klow,
!     *               nx,ny,nz)
!      ipsi4  = interp(  q(gr_psi4I),
!     *               fracx, fracy, fracz,
!     *               ilow,jlow,klow,
!     *               nx,ny,nz)
!
!      index = (jphi-1)*asf_shape(1) + (itheta-1)
!      !index = (jphi-1)*asf_ntheta + (itheta-1)
!      q(asf_rap4+index) = rpsi4
!      q(asf_iap4+index) = ipsi4
!
      if (ltrace) then
         write(*,*) 'grid_surface_comp: ********'
         write(*,*) 'grid_surface_comp: gi,hg     =',gi,hg
         write(*,9) 'grid_surface_comp: x,y,z     =',x,y,z
         write(*,8) 'grid_surface_comp: nx/y/z    =',nx,ny,nz
         write(*,8) 'grid_surface_comp: i/j/klow  =',ilow,jlow,klow
         write(*,9) 'grid_surface_comp: fracx/y/z =',fracx,fracy,fracz
         write(*,9) 'grid_surface_comp: xmin/ymin/zmin =',
     *                gr_minx(gi),gr_miny(gi),gr_minz(gi)
         write(*,9) 'grid_surface_comp: rps4       =',rpsi4

!       write(*,9) 'testA: ',q(gr_psi4r+(17-1)*nx*ny+(17-1)*nx+(17-1) ),
!     *      interp(q(gr_psi4r), 1.d0,1.d0,1.d0,
!     *        16,16,16, nx,ny,nz)
!       write(*,9) 'testb: ',q(gr_psi4r+(16-1)*nx*ny+(16-1)*nx+(16-1) ),
!     *      interp(q(gr_psi4r), 0.d0,0.d0,0.d0,
!     *        16,16,16, nx,ny,nz)
      end if

 8    format(A,5I10)
 9    format(A,5F10.4)

      return
      end    ! END: grid_surface_comp

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_bsurface_comp:                                                        cc
cc                   Takes as input the coordinate of a point located         cc
cc                   on the grid. It needs to compute the various             cc
cc                   surface functions at that point.                         cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_bsurface_comp(gi, x,y,z, itheta, jphi)
      implicit     none
      integer      gi, itheta, jphi
      real(kind=8) x,y,z
      include     'grid.inc'
      include     'surfaces.inc'
      integer      nx, ny, nz
      integer      i,j,k,index, l
      integer      ilow, jlow, klow
      real(kind=8) hg, rpsi4, ipsi4, fracx, fracy, fracz
      real(kind=8) interp
      external     interp
      !
      logical     ltrace
      parameter ( ltrace = .false. )

      return

      call load_pointers(gi)

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)

      ! Index of lower left point of interpolation cell:
      ilow  = INT( (x - gr_minx(gi)) / hg ) + 1
      jlow  = INT( (y - gr_miny(gi)) / hg ) + 1
      klow  = INT( (z - gr_minz(gi)) / hg ) + 1

      ! Fractional distance point is from lower left point:
      fracx = ( x - q( gr_x(gi) + (ilow-1) ) ) / hg
      fracy = ( y - q( gr_y(gi) + (jlow-1) ) ) / hg
      fracz = ( z - q( gr_z(gi) + (klow-1) ) ) / hg

      if (ilow .lt. 1) then
         write(*,*) 'grid_surface_comp: x too small ',ilow
         call my_exit('grid_surface_comp: X too small')
      end if
      if (fracx.lt.-0.01 .or.fracy.lt.-0.01.or.fracz.lt.-0.0001) then
         write(*,*) 'grid_surface_comp: fraction cannot be negative'
         write(*,*) 'grid_surface_comp: fracx/y/z =',fracx,fracy,fracz
         call my_exit('grid_surface_comp: fraction cannot be negative')
      end if

!      rpsi4 = interp(  q(gr_psi4R),
!     *               fracx, fracy, fracz,
!     *               ilow,jlow,klow,
!     *               nx,ny,nz)
!      ipsi4  = interp(  q(gr_psi4I),
!     *               fracx, fracy, fracz,
!     *               ilow,jlow,klow,
!     *               nx,ny,nz)
!
!!      index = (jphi-1)*bsf_shape(1) + (itheta-1)
!      !index = (jphi-1)*asf_ntheta + (itheta-1)
!      q(bsf_rbp4+index) = rpsi4
!      q(bsf_ibp4+index) = ipsi4
!
      if (ltrace) then
         write(*,*) 'grid_surface_comp: ********'
         write(*,*) 'grid_surface_comp: gi,hg     =',gi,hg
         write(*,9) 'grid_surface_comp: x,y,z     =',x,y,z
         write(*,8) 'grid_surface_comp: nx/y/z    =',nx,ny,nz
         write(*,8) 'grid_surface_comp: i/j/klow  =',ilow,jlow,klow
         write(*,9) 'grid_surface_comp: fracx/y/z =',fracx,fracy,fracz
         write(*,9) 'grid_surface_comp: xmin/ymin/zmin =',
     *                gr_minx(gi),gr_miny(gi),gr_minz(gi)
         write(*,9) 'grid_surface_comp: rps4       =',rpsi4

!       write(*,9) 'testA: ',q(gr_psi4r+(17-1)*nx*ny+(17-1)*nx+(17-1) ),
!     *      interp(q(gr_psi4r), 1.d0,1.d0,1.d0,
!     *        16,16,16, nx,ny,nz)
!       write(*,9) 'testb: ',q(gr_psi4r+(16-1)*nx*ny+(16-1)*nx+(16-1) ),
!     *      interp(q(gr_psi4r), 0.d0,0.d0,0.d0,
!     *        16,16,16, nx,ny,nz)
      end if

 8    format(A,5I10)
 9    format(A,5F10.4)

      return
      end    ! END: grid_surface_comp

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_csurface_comp:                                                        cc
cc                   Takes as input the coordinate of a point located         cc
cc                   on the grid. It needs to compute the various             cc
cc                   surface functions at that point.                         cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_csurface_comp(gi, x,y,z, itheta, jphi)
      implicit     none
      integer      gi, itheta, jphi
      real(kind=8) x,y,z
      include     'grid.inc'
      include     'surfaces.inc'
      integer      nx, ny, nz
      integer      i,j,k,index, l
      integer      ilow, jlow, klow
      real(kind=8) hg, rpsi4, ipsi4, fracx, fracy, fracz
      real(kind=8) interp
      external     interp
      !
      logical     ltrace
      parameter ( ltrace = .false. )

      return

      call load_pointers(gi)

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)

      ! Index of lower left point of interpolation cell:
      ilow  = INT( (x - gr_minx(gi)) / hg ) + 1
      jlow  = INT( (y - gr_miny(gi)) / hg ) + 1
      klow  = INT( (z - gr_minz(gi)) / hg ) + 1

      ! Fractional distance point is from lower left point:
      fracx = ( x - q( gr_x(gi) + (ilow-1) ) ) / hg
      fracy = ( y - q( gr_y(gi) + (jlow-1) ) ) / hg
      fracz = ( z - q( gr_z(gi) + (klow-1) ) ) / hg

      if (ilow .lt. 1) then
         write(*,*) 'grid_surface_comp: x too small ',ilow
         call my_exit('grid_surface_comp: X too small')
      end if
      if (fracx.lt.-0.01 .or.fracy.lt.-0.01.or.fracz.lt.-0.0001) then
         write(*,*) 'grid_surface_comp: fraction cannot be negative'
         write(*,*) 'grid_surface_comp: fracx/y/z =',fracx,fracy,fracz
         call my_exit('grid_surface_comp: fraction cannot be negative')
      end if

!      rpsi4 = interp(  q(gr_psi4R),
!     *               fracx, fracy, fracz,
!     *               ilow,jlow,klow,
!     *               nx,ny,nz)
!      ipsi4  = interp(  q(gr_psi4I),
!     *               fracx, fracy, fracz,
!     *               ilow,jlow,klow,
!     *               nx,ny,nz)
!
!      index = (jphi-1)*csf_shape(1) + (itheta-1)
!      !index = (jphi-1)*asf_ntheta + (itheta-1)
!      q(csf_rcp4+index) = rpsi4
!      q(csf_icp4+index) = ipsi4

      if (ltrace) then
         write(*,*) 'grid_surface_comp: ********'
         write(*,*) 'grid_surface_comp: gi,hg     =',gi,hg
         write(*,9) 'grid_surface_comp: x,y,z     =',x,y,z
         write(*,8) 'grid_surface_comp: nx/y/z    =',nx,ny,nz
         write(*,8) 'grid_surface_comp: i/j/klow  =',ilow,jlow,klow
         write(*,9) 'grid_surface_comp: fracx/y/z =',fracx,fracy,fracz
         write(*,9) 'grid_surface_comp: xmin/ymin/zmin =',
     *                gr_minx(gi),gr_miny(gi),gr_minz(gi)
         write(*,9) 'grid_surface_comp: rps4       =',rpsi4

!       write(*,9) 'testA: ',q(gr_psi4r+(17-1)*nx*ny+(17-1)*nx+(17-1) ),
!     *      interp(q(gr_psi4r), 1.d0,1.d0,1.d0,
!     *        16,16,16, nx,ny,nz)
!       write(*,9) 'testb: ',q(gr_psi4r+(16-1)*nx*ny+(16-1)*nx+(16-1) ),
!     *      interp(q(gr_psi4r), 0.d0,0.d0,0.d0,
!     *        16,16,16, nx,ny,nz)
      end if

 8    format(A,5I10)
 9    format(A,5F10.4)

      return
      end    ! END: grid_surface_comp
