Do NOT Edit this file...it has been copied here from /home2/steve/projects/had/src/hyperMHD
!
!
! grid_ell.f:  These are a collection of project-specific routines
!              which address the elliptic problem. If the project
!              has no elliptic equation to solve, these should just
!              be dummy routines (declared, but no body necessary).
!
!              The idea is that the call sequences to the "normal" 
!              lop(), get_resid(), and relax() routines need to be 
!              hidden from the infrastructure and kept secured in  
!              the project directories.                             
!
!
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_mg_trace:                                                            cc
cc            Output fields relevenat to the multigrid solver for tracing     cc
cc            and debugging.                                                  cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_mg_trace( gi, count )
      implicit    none
      integer     gi, count
      include    'grid.inc'
      include    'glob.inc'
      integer     nx,  ny,  nz, myint
      real*8      hg, time
      integer      proc_return_myid, myid, lev
      external     proc_return_myid
      logical     ltrace
      parameter ( ltrace = .false. )
      integer    i,j,k,index
      integer    nxc,nyc,nzc, kc,indexc
      character(2) tmps

      myid = proc_return_myid()

      call load_pointers(gi)

      if (ltrace) write(*,*) 'grid_mg_trace: On grid: ',gi, count

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)
      lev  = gr_level(gi)
      call int2str(lev,tmps)
      !time = 1.d0 * count
      ! Offset time w/r/t real time:
      myint = NINT(1000*gr_t(gi))
      time = 1.d0*count + 100000*myint

      !
      ! Output 2D field in the middle of the z-axis:
      !
      k    = (nz+1)/2
      index= (k-1)*nx*ny

      ! For the restricted fields:
      nxc    = NINT( 0.5d0*(nx+1) )
      nyc    = NINT( 0.5d0*(ny+1) )
      nzc    = NINT( 0.5d0*(nz+1) )
      kc     = (nzc+1)/2
      indexc = (kc-1)*nxc*nyc

c     call field_out2d( q(gr_myalpha+index),
c    *                  time,
c    *                  'MGmyalpha'//tmps,
c    *                  gr_minx(gi), gr_maxx(gi),
c    *                  gr_miny(gi), gr_maxy(gi),
c    *                  nx,ny, myid)

      return
      end    ! END: grid_mg_trace
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_get_resid:                                                           cc
cc            Compute the residual of the current solution and stick          cc
cc            it into the u_st1 arrays.                                       cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_get_resid( gi )
      implicit    none
      integer     gi
      include    'grid.inc'
      integer     nx,  ny,  nz
      integer     nxc, nyc, nzc
      real*8      hg, hc
      logical     ltrace
      parameter ( ltrace = .false. )
      integer    i,j,k,index

      call load_pointers(gi)

      if (ltrace) write(*,*) 'grid_get_resid: On grid: ',gi

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)

      nxc = NINT( 0.5d0*(nx+1) )
      nyc = NINT( 0.5d0*(ny+1) )
      nzc = NINT( 0.5d0*(nz+1) )
      hc  = 2.d0*hg

c     call get_resid(q(gr_psi_elliptic_st1),
c    *               q(gr_psi_elliptic),
c    *               q(gr_psi_elliptic_rhs),hg,nx,ny,nz)

      return
      end    ! END: grid_get_resid

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_lop:                                                                 cc
cc            Compute the results of the differential operator associated     cc
cc            with the elliptic problem acting on the set of restricted       cc
cc            fields residing in the u_st1 fields and *add* the result        cc
cc            to the u_st2 arrays.                                            cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_lop( gi )
      implicit    none
      integer     gi
      include    'grid.inc'
      integer     nx,  ny,  nz
      integer     nxc, nyc, nzc
      real*8      hg, hc
      logical     ltrace
      parameter ( ltrace = .false. )
      integer    i,j,k,index

      call load_pointers(gi)

      if (ltrace) write(*,*) 'grid_lop:    On grid: ',gi

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)

      nxc = NINT( 0.5d0*(nx+1) )
      nyc = NINT( 0.5d0*(ny+1) )
      nzc = NINT( 0.5d0*(nz+1) )
      hc  = 2.d0*hg

c     call lop( q(gr_psi_elliptic_st2),
c    *          q(gr_psi_elliptic_st1),hc,nxc,nyc,nzc)

      return
      end    ! END: grid_lop

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_ellrhs:                                                              cc
cc               Compute the RHS for the elliptic solve.                      cc
cc               If no elliptic solver in project, just leave empty.          cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_ellrhs( gridnum )
      implicit    none
      integer     gridnum
      include    'grid.inc'
      integer     nx, ny, nz
      real*8      hg
      logical     ltrace
      parameter ( ltrace = .false. )
      integer    i,j,k,index
      integer    indexip1, indexim1
      integer    indexjp1, indexjm1
      integer    indexkp1, indexkm1

      call load_pointers(gridnum)

      if (ltrace) write(*,*) 'grid_ellrhs:    On grid: ',gridnum

      nx   = gr_nx(gridnum)
      ny   = gr_ny(gridnum)
      nz   = gr_nz(gridnum)
      hg   = gr_h(gridnum)

      !
      ! Zero out rhs:
      !
c     call load_scal1D(q(gr_psi_elliptic_rhs),0.d0,nx*ny*nz)

      !
      ! Compute interior:
      !
      do k = 2, nz-1
         do j = 2, ny-1
            do i = 2, nx-1
               index    = (k-1)*ny*nx + (j-1)*nx + (i-1)
               !
               indexip1 = (k-1)*ny*nx + (j-1)*nx + (i  )
               indexim1 = (k-1)*ny*nx + (j-1)*nx + (i-2)
               !
               indexjp1 = (k-1)*ny*nx + (j  )*nx + (i-1)
               indexjm1 = (k-1)*ny*nx + (j-2)*nx + (i-1)
               !
               indexkp1 = (k  )*ny*nx + (j-1)*nx + (i-1)
               indexkm1 = (k-2)*ny*nx + (j-1)*nx + (i-1)
               !
c              q(gr_psi_elliptic_rhs+index) = 
c    *                (q(gr_bx_np1+indexip1) - q(gr_bx_np1+indexim1))/hg
c    *              + (q(gr_by_np1+indexjp1) - q(gr_by_np1+indexjm1))/hg
c    *              + (q(gr_bz_np1+indexkp1) - q(gr_bz_np1+indexkm1))/hg
c    *                (q(gr_bx    +indexip1) - q(gr_bx    +indexim1))/hg
c    *              + (q(gr_by    +indexjp1) - q(gr_by    +indexjm1))/hg
c    *              + (q(gr_bz    +indexkp1) - q(gr_bz    +indexkm1))/hg
            end do
         end do
      end do

      !
      ! For debugging only set to a Gaussian:
      !
c     call gaussian3d( q(gr_psi_elliptic_rhs), hg, gr_minx(gridnum),
c    *     gr_miny(gridnum), gr_minz(gridnum),nx,ny,nz,
c    *     1.d0,1.d0,1.d0,0.d0,0.1d0,0.d0,0.d0, 0.d0)

c     if (ltrace) call field_dump_info(q(gr_psi_elliptic_rhs),nx,ny,nz)

      return
      end    ! END: grid_ellrhs

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_relax:                                                               cc
cc            Apply relaxation routine to field(s) on a given grid.           cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_relax(gi)
      implicit      none
      integer       gi
      include      'grid.inc'
      include      'glob.inc'

      real(kind=8)   relax
      external       relax
      real(kind=8)   resid

      logical     ltrace
      parameter ( ltrace  = .false. )


      if (ltrace) then
         write(*,*) 'grid_relax: Relaxing grid: ',gi
      end if

      call load_pointers(gi)

c     resid = relax(q(gr_psi_elliptic),q(gr_psi_elliptic_rhs),q(gr_chr),
c    *                gr_h(gi),gr_nx(gi),gr_ny(gi),gr_nz(gi))

      !
      ! For output and debugging purposes only:
      !
!     call get_resid(q(gr_phi_x),q(gr_phi),q(gr_rho),
!    *                gr_h(gi),gr_nx(gi),gr_ny(gi),gr_nz(gi))

      if (ltrace) then
         write(*,*) 'grid_relax: Finished with resid: ',resid
      end if

      return
      end      ! END: grid_relax

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_return_resnrm:  Get norm of residual on a given grid (for debugging):cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      double precision function grid_return_resnrm(gi)
      implicit      none
      integer       gi
      include      'grid.inc'
      integer       nx, ny, nz

      real(kind=8)  h
      real(kind=8)  get_resnrm
      external      get_resnrm
      real(kind=8)  grid_return_h
      external      grid_return_h
      logical       grid_is_local
      external      grid_is_local

      logical     ltrace
      parameter ( ltrace = .false. )


      call grid_get_dims(gi,nx,ny,nz)
      h = grid_return_h(gi)

      call load_pointers(gi)

      if (.not. grid_is_local(gi)) then
         write(*,*) 'grid_return_resnrm: Only works on local grids'
      end if

      grid_return_resnrm = 0.d0
c     grid_return_resnrm = get_resnrm(q(gr_psi_elliptic),
c    *                          q(gr_psi_elliptic_rhs),h,nx,ny,nz)

      return
      end      ! END: grid_return_resnrm

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_ellpost:                                                             cc
cc               Things to do after an elliptic solve.                        cc
cc               For MHD project, this is where we clean the B field.         cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_ellpost( gridnum )
      implicit    none
      integer     gridnum
      include    'grid.inc'
      include    'param.inc'
      integer     nx, ny, nz
      real*8      hg
      logical     ltrace
      parameter ( ltrace = .true. )
      integer    i,j,k,index
      integer    indexip1, indexim1
      integer    indexjp1, indexjm1
      integer    indexkp1, indexkm1

      call load_pointers(gridnum)

      if (ltrace) write(*,*) 'grid_ellpost:   On grid: ',gridnum

      nx   = gr_nx(gridnum)
      ny   = gr_ny(gridnum)
      nz   = gr_nz(gridnum)
      hg   = gr_h(gridnum)


      return
      end    ! END: grid_ellpost

